#include "hmm.h"

#include <stdlib.h>
#include <assert.h>
#include <values.h>
#include <math.h>

Sequence *sequence_create(int T, int v) {
	Sequence *seq;
	int i;
	int *p;

	if((seq = malloc(sizeof(*seq)+v*sizeof(int*))) == NULL) return NULL;
	if((p = malloc(T*v*sizeof(*p))) == NULL) return NULL;
	seq->T = T;
	seq->v = v;
	for(i=0; i<v; i++) {
		seq->o[i] = p;
		p += T;
	}
	return seq;
}

void sequence_destroy(Sequence *seq) {
	free(seq->o[0]);
	free(seq);
}

HMM *hmm_create(int n, int v, int m[], int lr) {
	HMM *hmm;
	int i;

	if((hmm = malloc(sizeof(*hmm)+v*sizeof(Matrix*))) == NULL) return NULL;
	hmm->n = n;
	hmm->v = v;
	hmm->pi = vector_create(n);
	hmm->lr = lr;
	hmm->A = matrix_create(n, n);
	for(i=0; i<v; i++) hmm->B[i] = matrix_create(n, m[i]);
	return hmm;
}

HMM *hmm_clone(HMM *old) {
	int i;
	int m[old->v];
	HMM *hmm;

	for(i=0; i<old->v; i++) m[i] = old->B[i]->n;
	hmm = hmm_create(old->n, old->v, m, old->lr);
	vector_copy(hmm->pi, old->pi);
	matrix_copy(hmm->A, old->A);
	for(i=0; i<old->v; i++) matrix_copy(hmm->B[i], old->B[i]);
	return hmm;
}

void hmm_dump(HMM *hmm, FILE *file) {
	int i;

	fputc(hmm->n & 0xff, file);
	fputc(hmm->v & 0xff, file);
	fputc(hmm->lr & 0xff, file);
	for(i=0; i<hmm->v; i++) {
		fputc(hmm->B[i]->n, file);
	}
	fwrite(hmm->pi->x, sizeof(double), hmm->n, file);
	fwrite(hmm->A->x[0], sizeof(double), hmm->n*hmm->n, file);
	for(i=0; i<hmm->v; i++) {
		fwrite(hmm->B[i]->x[0], sizeof(double),
			hmm->B[i]->m*hmm->B[i]->n, file);
	}
}

HMM *hmm_load(FILE *file) {
	int i, n, v, lr;
	HMM *hmm;

	n = fgetc(file);
	v = fgetc(file);
	lr = fgetc(file);

	int m[v];

	for(i=0; i<v; i++) {
		m[i] = fgetc(file);
	}

	hmm = hmm_create(n, v, m, lr);
	fread(hmm->pi->x, sizeof(double), n, file);
	fread(hmm->A->x[0], sizeof(double), n*n, file);
	for(i=0; i<v; i++) {
		fread(hmm->B[i]->x[0], sizeof(double), n*m[i], file);
	}

	return hmm;
}

void hmm_destroy(HMM *hmm) {
	int i;

	vector_destroy(hmm->pi);
	matrix_destroy(hmm->A);
	for(i=0; i<hmm->v; i++) matrix_destroy(hmm->B[i]);
	free(hmm);
}

void hmm_randomize(HMM *hmm) {
	int i,j,k;

	if(hmm->lr) {
		hmm->pi->x[0] = 1.0;
		for(i=1; i<hmm->n; i++) hmm->pi->x[i] = 0.0;
	} else {
		for(i=0; i<hmm->n; i++)
			hmm->pi->x[i] = 1.0 + ((double)rand())/((double)RAND_MAX);
		vector_normalize(hmm->pi);
	}

	if(hmm->lr) {
		for(i=0; i<hmm->n; i++) {
			for(j=0; j<i; j++)
				hmm->A->x[i][j] = 0.0;
			for(; j<i+hmm->lr && j<hmm->n; j++)
				hmm->A->x[i][j] = 1.0 + ((double)rand())/((double)RAND_MAX);
			for(; j<hmm->n; j++)
				hmm->A->x[i][j] = 0.0;
		}
	} else {
		for(i=0; i<hmm->n; i++) {
			for(j=0; j<hmm->n; j++)
				hmm->A->x[i][j] = 1.0 + ((double)rand())/((double)RAND_MAX);
		}
	}

	for(k=0; k<hmm->v; k++) {
		for(i=0; i<hmm->B[k]->m; i++) {
			for(j=0; j<hmm->B[k]->n; j++)
				hmm->B[k]->x[i][j] = 1.0+((double)rand())/((double)RAND_MAX);
		}
		matrix_normalize_rows(hmm->B[k]);
	}
	matrix_normalize_rows(hmm->A);
}

static void hmm_alpha_c(HMM *hmm, int *o[], int T, Matrix *a, Vector *c) {
	Vector *u;
	int i,t;

	assert(a->n == T && a->m == hmm->n && c->n == T);

	u = vector_create(hmm->n);
	vector_mul_each_col(u, hmm->pi, hmm->B[0], o[0][0]);
	for(i=1; i<hmm->v; i++)
		vector_mul_each_col(u, u, hmm->B[i], o[i][0]);
	c->x[0] = vector_sum(u);
	c->x[0] = (c->x[0] == 0.0)? 0.0 : 1.0 / c->x[0];
	vector_scale(u, c->x[0]);
	matrix_col_from_vector(a, 0, u);

	for(t=1; t<T; t++) {
		vector_mul_matrix(u, u, hmm->A);
		vector_mul_each_col(u, u, hmm->B[0], o[0][t]);
		for(i=1; i<hmm->v; i++)
			vector_mul_each_col(u, u, hmm->B[i], o[i][t]);
		c->x[t] = vector_sum(u);
		c->x[t] = (c->x[t] == 0.0)? 0.0 : 1.0 / c->x[t];
		vector_scale(u, c->x[t]);
		matrix_col_from_vector(a, t, u);
	}
	vector_destroy(u);
}

static void hmm_beta(HMM *hmm, int *o[], int T, Matrix *b, Vector *c) {
	Vector *u;
	int i,k,t;

	assert(b->m == hmm->n && b->n == T && c->n == T);

	u = vector_create(hmm->n);
	for(i=0; i<u->n; i++) u->x[i] = c->x[T-1];
	matrix_col_from_vector(b, T-1, u);
	for(t=T-2; t>=0; t--) {
		vector_mul_each_col(u, u, hmm->B[0], o[0][t+1]);
		for(k=1; k<hmm->v; k++)
			vector_mul_each_col(u, u, hmm->B[k], o[k][t+1]);
		matrix_mul_vector(u, hmm->A, u);
		vector_scale(u, c->x[t]);
		matrix_col_from_vector(b, t, u);
	}
	vector_destroy(u);
}

static void hmm_gamma(HMM *hmm, int *o[], int T, Matrix *a, Matrix *b,
Matrix *g, Matrix *dg[]) {
	int i,j,k,t;
	double denom, scale, x;

	assert(a->m == hmm->n && a->n == T);
	assert(b->m == hmm->n && b->n == T);
	assert(g->m == hmm->n && g->n == T);

	for(t=0; t<T-1; t++) {
		denom = 0.0;
		for(i=0; i<hmm->n; i++) {
			for(j=0; j<hmm->n; j++) {
				x = a->x[i][t] * hmm->A->x[i][j] * b->x[j][t+1];
				for(k=0; k<hmm->v; k++)
					x *= hmm->B[k]->x[j][o[k][t+1]];
				denom += x;
			}
		}
		scale = (denom == 0.0)? 0.0 : 1.0 / denom;
		for(i=0; i<hmm->n; i++) {
			g->x[i][t] = 0.0;
			for(j=0; j<hmm->n; j++) {
				x = a->x[i][t] * hmm->A->x[i][j] * b->x[j][t+1] * scale;
				for(k=0; k<hmm->v; k++)
					x *= hmm->B[k]->x[j][o[k][t+1]];
				g->x[i][t] += (dg[t]->x[i][j] = x);
			}
		}
	}
	for(i=0; i<hmm->n; i++) g->x[i][T-1] = a->x[i][T-1];
}

static void hmm_update(HMM *hmm, int *o[], size_t T, Matrix *a, Matrix *b,
Vector *c, Matrix *g, Matrix *dg[]) {
	int i,j,k,t;
	double numer, denom, scale;

	assert(a->m == hmm->n && a->n == T);
	assert(b->m == hmm->n && b->n == T);
	assert(c->n == T);
	assert(g->m == hmm->n && g->n == T);
#ifndef NDEBUG
	for(i=0; i<T-1; i++) assert(dg[i]->n == hmm->n && dg[i]->m == hmm->n);
#endif

	if(!hmm->lr) {
		vector_add_col(hmm->pi, hmm->pi, g, 0);
	}

	for(i=0; i<hmm->n; i++) {
		denom = 0.0;
		for(t=0; t<T-1; t++) denom += g->x[i][t];
		scale = (denom == 0.0)? 0.0 : 1.0 / denom;
		for(j=0; j<hmm->n; j++) {
			numer = 0.0;
			for(t=0; t<T-1; t++) {
				numer += dg[t]->x[i][j];
			}
			hmm->A->x[i][j] += numer * scale;
		}
	}

	for(k=0; k<hmm->v; k++) {
		for(i=0; i<hmm->n; i++) {
			denom = matrix_sum_row(g, i);
			scale = (denom == 0.0)? 0.0 : 1.0 / denom;
			for(j=0; j<hmm->B[k]->n; j++) {
				numer = 0.0;
				for(t=0; t<T; t++) {
					if(o[k][t] == j) numer += g->x[i][t];
				}
				hmm->B[k]->x[i][j] += numer * scale;
			}
		}
	}
}

double hmm_log_p(Vector *c) {
	double lp = 0.0;
	int i;

	for(i=0; i<c->n; i++) lp -= log(c->x[i]);
	return lp;
}

HMM *hmm_train(HMM *old, Sequence *seq[], int n) {
	int T = seq[0]->T;
	int v = seq[0]->v;
	Matrix *a, *b, *g;
	Matrix *dg[T-1];
	Vector *c;
	HMM *hmm;
	int i;
	double old_log_p, log_p = -MAXDOUBLE;

	a = matrix_create(old->n, T);
	b = matrix_create(old->n, T);
	c = vector_create(T);
	g = matrix_create(old->n, T);
	for(i=0; i<T-1; i++) dg[i] = matrix_create(old->n, old->n);

	do {
		hmm = hmm_clone(old);

		old_log_p = log_p;
		log_p = 0.0;
		for(i=0; i<n; i++) {
			hmm_alpha_c(old, seq[i]->o, T, a, c);
			hmm_beta(old, seq[i]->o, T, b, c);
			hmm_gamma(old, seq[i]->o, T, a, b, g, dg);
			hmm_update(hmm, seq[i]->o, T, a, b, c, g, dg);
			log_p += hmm_log_p(c);
		}

		vector_normalize(hmm->pi);
		matrix_normalize_rows(hmm->A);
		for(i=0; i<v; i++) matrix_normalize_rows(hmm->B[i]);

		hmm_destroy(old);
		old = hmm;
//		printf("%g -> %g\n", old_log_p, log_p);
	} while((old_log_p - log_p) / log_p > 1.0e-3);

	matrix_destroy(a);
	matrix_destroy(b);
	vector_destroy(c);
	matrix_destroy(g);
	for(i=0; i<T-1; i++) matrix_destroy(dg[i]);

	return hmm;
}

double hmm_compare(HMM *hmm, Sequence *seq) {
	int T = seq->T;
	int **o = seq->o;
	Vector *u;
	int i,t;
	double lp = 0.0;
	double c;

	u = vector_create(hmm->n);
	vector_mul_each_col(u, hmm->pi, hmm->B[0], o[0][0]);
	for(i=1; i<hmm->v; i++)
		vector_mul_each_col(u, u, hmm->B[i], o[i][0]);
	lp = log(c = vector_sum(u));
	c = 1.0 / c;
	vector_scale(u, c);

	for(t=1; t<T; t++) {
		vector_mul_matrix(u, u, hmm->A);
		vector_mul_each_col(u, u, hmm->B[0], o[0][t]);
		for(i=1; i<hmm->v; i++)
			vector_mul_each_col(u, u, hmm->B[i], o[i][t]);
		lp += log(c = vector_sum(u));
		c = 1.0 / c;
		vector_scale(u, c);
	}
	vector_destroy(u);

	return lp;
}


