#include "hmm.h"
#include "xi.h"
#include "language.h"
#include "kandie.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <values.h>
#include <assert.h>

/* If you edit this, you might want to edit VARIABLES in kandie.h
 * Remmeber to fill out the zero items! Better safe than sorry. */
int res[] = { 64, 0, 0, 0, 0, 0, 0, 0 };

static enum { MODE_TRAIN, MODE_RUN } mode;
static Character **trainchar = NULL;
static int trainchar_len = 0;
static Language *language;
static char *stroke_filename = NULL;

static int training_step(void) {
	FILE *file;

	if(trainchar_len == 0) {
		free(trainchar);
		fprintf(stderr, "Training complete. Computing HMMs... ");
		fflush(stderr);
		language_train(language);
		fprintf(stderr, "done\n");
		if((file = fopen(stroke_filename, "w")) == NULL) {
			fprintf(stderr, "Can not open file for writing: \"%s\"\n",
				stroke_filename);
			return 1;
		}
		language_dump(language, file);
		fclose(file);
		fprintf(stderr, "Training done, exiting.\n");
		return 1;
	} else {
		fprintf(stderr, "%d items left. Write: \"%s\".\n",
			trainchar_len, trainchar[trainchar_len-1]->str);
	}
	return 0;
}

static int process_sequence(Sequence *seq) {
	static Sequence *seq_buf[MAX_STROKES];
	static int seq_buf_len = 0;

	int i;
	Character *character;

	if(seq == NULL) {
		switch(mode) {
		case MODE_TRAIN:
			assert(trainchar_len > 0);
			if(trainchar[trainchar_len-1]->strokes != seq_buf_len) {
				fprintf(stderr, "Invalid number of strokes.\n");
				seq_buf_len = 0;
				return 0;
			}
			character_add_example(language, trainchar[trainchar_len-1],
				seq_buf, seq_buf_len);
			seq_buf_len = 0;
			trainchar_len--;
			return training_step();
		case MODE_RUN:
			if(seq_buf_len == 0) {
				putchar(' ');
			} else {
				character = language_select(language, seq_buf, seq_buf_len);
				if(character != NULL) {
					fputs(character->str, stdout);
				}
			}
			fflush(stdout);
			for(i=0; i<seq_buf_len; i++) sequence_destroy(seq_buf[i]);
			seq_buf_len = 0;
			return 0;
		default:
			break;
		}
		return 0;
	}

	switch(mode) {
		case MODE_TRAIN:
		case MODE_RUN:
			if(seq_buf_len >= MAX_STROKES) {
				fprintf(stderr, "Stroke count limit exceeded.\n");
				for(i=0; i<seq_buf_len; i++)
					sequence_destroy(seq_buf[i]);
				seq_buf_len = 0;
				return 0;
			}
			seq_buf[seq_buf_len++] = seq;
			break;
		default:
			break;
	}
	return 0;
}

int main(int argc, char *argv[]) {
	int c, i, n;
	char *def_filename = NULL, *train_filename = NULL;
	FILE *file;

	mode = MODE_RUN;
	for(;;) {
		if((c = getopt(argc, argv, "d:hs:t:")) == -1) break;
		switch(c) {
			case 'd':
				def_filename = optarg;
				break;
			case 'h':
				fprintf(stderr,
	"Usage: kandie [-h] -d filename -s filename -t filename\n"
	" -h             Show this message\n"
	" -t filename    Enter training mode, using the given file\n"
	" -d filename    Specify language definition file\n"
	" -s filename    Specify stroke data file\n\n");
				return 0;
			case 's':
				stroke_filename = optarg;
				break;
			case 't':
				train_filename = optarg;
				mode = MODE_TRAIN;
				break;
			default:
				fprintf(stderr, "Unknown option. Try -h for help.\n");
				return 1;
		}
	}

	if(optind < argc) {
		fprintf(stderr, "Non-option arguments.\n");
		return 1;
	}

	if(def_filename == NULL || stroke_filename == NULL) {
		fprintf(stderr, "Definition and stroke files must be specified.\n");
		return 1;
	}

	if((file = fopen(def_filename, "r")) == NULL) {
		fprintf(stderr, "Can not open file: \"%s\"\n", def_filename);
		return 1;
	}
	if((language = language_load(file)) == NULL) {
		fprintf(stderr,
			"Can not load language definition from file: \"%s\"\n",
			def_filename);
		return 1;
	}
	fclose(file);

	if(mode == MODE_RUN) {
		if((file = fopen(stroke_filename, "r")) == NULL) {
			fprintf(stderr, "Can not open file: \"%s\"\n", stroke_filename);
			return 1;
		}
		if(language_load_strokes(language, file)) {
			fprintf(stderr,
				"Can not load stroke data from file: \"%s\"\n",
				stroke_filename);
			return 1;
		}
		fclose(file);
	} else {
		language_init_strokes(language);
		if((file = fopen(train_filename, "r")) == NULL) {
			fprintf(stderr, "Can not load training file: \"%s\"\n",
				train_filename);
			return 1;
		}
		if(fscanf(file, "%d", &trainchar_len) != 1) {
			fprintf(stderr, "Expected training set length.\n");
			return 1;
		}
		trainchar = malloc(trainchar_len*sizeof(*trainchar));
		for(i=trainchar_len-1; i>=0; i--) {
			if(fscanf(file, "%d", &n) != 1) {
				fprintf(stderr, "Expected training character numberr.\n");
				return 1;
			}
			trainchar[i] = language->character[n];
		}
		fclose(file);
		training_step();
	}

	xi_init("stylus", process_sequence, SAMPLES, res);

	return 0;
}

