#include "kandie.h"
#include "hmm.h"
#include "language.h"

#include <stdlib.h>
#include <string.h>
#include <values.h>
#include <assert.h>

Stroke *stroke_create(void) {
	Stroke *stroke;

	if((stroke = malloc(sizeof(*stroke)+32*sizeof(Sequence*))) == NULL)
		return NULL;
	stroke->seq_alloc = 32;
	stroke->n = 0;
	stroke->hmm = hmm_create(HIDDEN, VARIABLES, res, JUMPS);
	hmm_randomize(stroke->hmm);

	return stroke;
}

void stroke_destroy(Stroke *stroke) {
	int i;

	for(i=0; i<stroke->n; i++) sequence_destroy(stroke->seq[i]);
}

Stroke *stroke_add_sequence(Stroke *stroke, Sequence *seq) {
	if(stroke->n >= stroke->seq_alloc) {
		stroke->seq_alloc *= 2;
		stroke = realloc(stroke,
			sizeof(*stroke)+stroke->seq_alloc*sizeof(Sequence*));
		if(stroke == NULL) return NULL;
	}
	stroke->seq[stroke->n++] = seq;
	return stroke;
}

Character *character_create(Language *language, int strokes, int stroke_nr[],
char *str) {
	Character *character;

	if((character =
		malloc(sizeof(*character)+strokes*sizeof(int))) == NULL)
		return NULL;
	character->strokes = strokes;
	character->str = str;
	memcpy(character->stroke_nr, stroke_nr, strokes*sizeof(int));

	return character;
}

void character_destroy(Character *character) {
	free(character->str);
	free(character);
}

double character_compare(Language *language, Character *character,
Sequence *seq[], int n) {
	int i;
	double lp = 0.0;
	Stroke *stroke;

	assert(character->strokes == n);
	for(i=0; i<character->strokes; i++) {
		stroke = language->stroke[character->stroke_nr[i]];
		lp += hmm_compare(stroke->hmm, seq[i]);
	}
	return lp;
}

void character_add_example(Language *language, Character *character,
Sequence *seq[], int n) {
	int i;
	Stroke *stroke;
	for(i=0; i<character->strokes; i++) {
		stroke = language->stroke[character->stroke_nr[i]];
		stroke = stroke_add_sequence(stroke, seq[i]);
		language->stroke[character->stroke_nr[i]] = stroke;
	}
}

Language *language_create(void) {
	Language *language;

	if((language = malloc(sizeof(*language))) == NULL) return NULL;
	language->strokes = 0;
	language->stroke = NULL;
	language->characters = 0;
	language->character_alloc = 32;
	if((language->character =
		malloc(language->character_alloc*sizeof(Character*))) == NULL)
		return NULL;
	return language;
}

void language_dump(Language *language, FILE *file) {
	int i,j,k;
	int *x;
	Stroke *stroke;

	fputc(language->strokes, file);
	for(i=0; i<language->strokes; i++) {
		stroke = language->stroke[i];
		fputc(stroke->n, file);
		for(j=0; j<stroke->n; j++) {
			fputc(stroke->seq[j]->T & 0xff, file);
			fputc((stroke->seq[j]->T >> 8) & 0xff, file);
			fputc(stroke->seq[j]->v & 0xff, file);
			x = stroke->seq[j]->o[0];
			for(k=0; k<stroke->seq[j]->T*stroke->seq[j]->v; k++) {
				fputc(x[k] & 0xff, file);
				fputc((x[k] >> 16) & 0xff, file);
			}
		}
		hmm_dump(stroke->hmm, file);
	}
}

void language_init_strokes(Language *language) {
	int i;

	for(i=0; i<language->strokes; i++) {
		language->stroke[i] = stroke_create();
	}
}

int language_load_strokes(Language *language, FILE *file) {
	int i, j, k, n, T, v, strokes;
	int *x;
	Stroke *stroke;
	Sequence *seq;

	strokes = fgetc(file);
	assert(language->strokes == strokes);
	for(i=0; i<strokes; i++) {
		n = fgetc(file);
		if((stroke = malloc(sizeof(*stroke)+n*sizeof(Sequence*))) == NULL)
			return 1;
		stroke->n = n;
		stroke->seq_alloc = n;
		for(j=0; j<n; j++) {
			T = fgetc(file);
			T += fgetc(file) << 8;
			v = fgetc(file);
			seq = sequence_create(T, v);
			x = seq->o[0];
			for(k=0; k<T*v; k++) {
				x[k] = fgetc(file);
				x[k] += fgetc(file) << 8;
			}
			stroke = stroke_add_sequence(stroke, seq);
		}
		stroke->hmm = hmm_load(file);
		language->stroke[i] = stroke;
	}
	return 0;
}

Language *language_load(FILE *file) {
	char str[0x20];
	int i, j, n, strokes, stroke_count;
	int stroke_nr[MAX_STROKES];
	Character *character;
	Language *language;

	if((language = language_create()) == NULL) return NULL;
	if(fscanf(file, "%d %d", &n, &stroke_count) != 2) {
		fprintf(stderr, "Expected character count and number of strokes.\n");
		return NULL;
	}
	if(stroke_count <= 0 || stroke_count > 0xffff) {
		fprintf(stderr, "Error: Stroke count is %d.\n", stroke_count);
		return NULL;
	}
	language->strokes = stroke_count;
	language->stroke = malloc(stroke_count*sizeof(*(language->stroke)));
	if(language->stroke == NULL) return NULL;

	for(i=0; i<n; i++) {
		if(fscanf(file, "%31s %d", str, &strokes) != 2) {
			fprintf(stderr, "Expected character + stroke count.\n");
			return NULL;
		}
		for(j=0; j<strokes; j++) {
			if(fscanf(file, "%d", &stroke_nr[j]) != 1) {
				fprintf(stderr, "Expected stroke number in \"%s\"\n", str);
				return NULL;
			}
		}
		character = character_create(
			language, strokes, stroke_nr, strdup(str));
		language_add_character(language, character);
	}
	return language;
}

void language_add_character(Language *language, Character *character) {
	if(language->characters >= language->character_alloc) {
		language->character_alloc *= 2;
		language->character = realloc(language->character,
			language->character_alloc*sizeof(Character*));
	}
	language->character[language->characters++] = character;
}

Character *language_select(Language *language, Sequence *seq[], int n) {
	Character *character, *best_character = NULL;
	double best_lp = -MAXDOUBLE, lp;
	int i;

	for(i=0; i<language->characters; i++) {
		character = language->character[i];
		if(character->strokes == n) {
			lp = character_compare(language, character, seq, n);
			if(lp > best_lp) {
				best_lp = lp;
				best_character = character;
			}
		}
	}
	return best_character;
}

void language_train(Language *language) {
	int i;
	Stroke *stroke;

	for(i=0; i<language->strokes; i++) {
		stroke = language->stroke[i];
		if(stroke->n > 0)
			stroke->hmm = hmm_train(stroke->hmm, stroke->seq, stroke->n);
	}
}


