#include "matrix.h"

#include <stdlib.h>
#include <string.h>
#include <assert.h>

Matrix *matrix_create(int m, int n) {
	Matrix *matrix;
	double *x;
	int i;

	if((matrix = malloc(sizeof(*matrix)+m*sizeof(double*))) == NULL)
		return NULL;
	if((x = malloc(m*n*sizeof(double))) == NULL) return NULL;
	matrix->n = n;
	matrix->m = m;
	for(i=0; i<m; i++) {
		matrix->x[i] = x;
		x += n;
	}
	return matrix;
}

Vector *vector_create(int n) {
	Vector *v;

	if((v = malloc(sizeof(*v)+n*sizeof(double))) == NULL) return NULL;
	v->n = n;
	return v;
}

void matrix_destroy(Matrix *m) {
	free(m->x[0]);
	free(m);
}

void vector_destroy(Vector *v) {
	free(v);
}

void vector_copy(Vector *d, Vector *v) {
	assert(d->n == v->n);
	memcpy(d->x, v->x, d->n*sizeof(double));
}

void matrix_copy(Matrix *d, Matrix *m) {
	assert(d->m == m->m && d->n == m->n);
	memcpy(d->x[0], m->x[0], d->n*d->m*sizeof(double));
}

void matrix_col_from_vector(Matrix *m, int j, Vector *v) {
	int i;

	assert(m->m == v->n && j >= 0 && j < m->n);
	for(i=0; i<m->m; i++) m->x[i][j] = v->x[i];
}

void vector_normalize(Vector *v) {
	int i;
	double s;

	s = v->x[0];
	for(i=1; i<v->n; i++) s += v->x[i];
	s = (s == 0.0)? 0.0 : 1.0 / s;
	for(i=0; i<v->n; i++) v->x[i] *= s;
}

static void matrix_normalize_row(Matrix *m, int i) {
	int j;
	double s;

	assert(i >= 0 && i < m->m);
	s = m->x[i][0];
	for(j=1; j<m->n; j++) s += m->x[i][j];
	if(s == 0.0) s = 0.0; else s = 1.0 / s;
	for(j=0; j<m->n; j++) m->x[i][j] *= s;
}

void matrix_normalize_rows(Matrix *m) {
	int i;
	for(i=0; i<m->m; i++) matrix_normalize_row(m, i);
}

void vector_mul_matrix(Vector *d, Vector *v, Matrix *m) {
	int i,j;
	double x[m->n];

	assert(d->n == m->n && v->n == m->m);
	for(j=0; j<m->n; j++) {
		x[j] = v->x[0] * m->x[0][j];
		for(i=1; i<m->m; i++)
			x[j] += v->x[i] * m->x[i][j];
	}
	memcpy(d->x, x, m->n*sizeof(double));
}

void vector_mul_each_col(Vector *d, Vector *v, Matrix *m, int j) {
	int i;

	assert(d->n == v->n && v->n == m->m && j >= 0 && j < m->n);
	for(i=0; i<v->n; i++) d->x[i] = v->x[i] * m->x[i][j];
}

void vector_add_col(Vector *d, Vector *v, Matrix *m, int j) {
	int i;

	assert(d->n == v->n && v->n == m->m && j >= 0 && j < m->n);
	for(i=0; i<v->n; i++) d->x[i] = v->x[i] + m->x[i][j];
}

void matrix_mul_vector(Vector *d, Matrix *m, Vector *v) {
	int i,j;
	double x[m->m];

	assert(d->n == m->m && v->n == m->n);
	for(i=0; i<m->m; i++) {
		x[i] = v->x[0] * m->x[i][0];
		for(j=1; j<m->n; j++)
			x[i] += v->x[j] * m->x[i][j];
	}
	memcpy(d->x, x, m->m*sizeof(double));
}

double vector_sum(Vector *v) {
	int i;
	double x;

	x = v->x[0];
	for(i=1; i<v->n; i++) x += v->x[i];
	return x;
}

double matrix_sum_col(Matrix *m, int j) {
	int i;
	double x;

	assert(j >= 0 && j < m->n);
	x = m->x[0][j];
	for(i=1; i<m->m; i++) x += m->x[i][j];
	return x;
}

double matrix_sum_row(Matrix *m, int i) {
	int j;
	double x;

	assert(i >= 0 && i < m->m);
	x = m->x[i][0];
	for(j=1; j<m->n; j++) x += m->x[i][j];
	return x;
}

void vector_scale(Vector *v, double k) {
	int i;

	for(i=0; i<v->n; i++) v->x[i] *= k;
}

void vector_print(FILE *file, Vector *v) {
	int i;

	for(i=0; i<v->n; i++) fprintf(file, "\t%.04f", v->x[i]);
	fputc('\n', file);
}

void matrix_print(FILE *file, Matrix *m) {
	int i,j;

	for(i=0; i<m->m; i++) {
		for(j=0; j<m->n; j++) {
			fprintf(file, "\t%.04f", m->x[i][j]);
		}
		fputc('\n', file);
	}
}

