#include "hmm.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <time.h>
#include <math.h>
#include <values.h>

#include <X11/Xlib.h>
#include <X11/extensions/XInput.h>
#include <X11/extensions/XIproto.h>

typedef struct {
	int x,y;	/* Position */
	int p;		/* Pressure */
	int tx,ty;	/* Tilt */
	double t;	/* Time in microseconds */
} Sample;

enum {
	INPUTEVENT_BTN_PRESS,
	INPUTEVENT_BTN_RELEASE,
	INPUTEVENT_MOTION_NOTIFY,

	INPUTEVENT_MAX
};

static XDeviceInfo *dev_list = NULL;
static int dev_list_cnt = 0;

static double move_treshold = 16.0;

static int input_event[INPUTEVENT_MAX];

/* If this is non-zero, pressure data rather than BTN_PRESS/BTN_RELEASE events
 * are used to check if the pen is down or up. */
static int use_pressure = 1;
static int pressure_treshold[2] = {
	64,		/* Pressure to enable "pen down" */
	48 } ;	/* Pressure to disable "pen down" */

static int error_handler(Display *disp, XErrorEvent *event) {
	char buf[0x80];

	XGetErrorText(disp, event->error_code, buf, sizeof(buf));
	fprintf(stderr, "X error: %d %s\n", event->error_code, buf);
	return 0;
}

static Display *get_display(void) {
	Display *disp;
	int major,fev,fer;

	if((disp = XOpenDisplay(NULL)) == NULL) {
		fprintf(stderr, "Failed to connect to X server.\n");
		return NULL;
	}

	XSetErrorHandler(error_handler);
	XSynchronize(disp, 1);

	if(!XQueryExtension(disp, INAME, &major, &fev, &fer)) {
		fprintf(stderr, "Server does not support XInput.\n");
		XCloseDisplay(disp);
		return NULL;
	}
	return disp;
}

static XDeviceInfo *get_device(Display *disp, const char *name) {
	int i;

	if(dev_list == NULL) {
		dev_list = (XDeviceInfoPtr) XListInputDevices(
			disp, &dev_list_cnt);
		if(dev_list == NULL) {
			fprintf(stderr, "Failed to get input device list.\n");
			return NULL;
		}
	}

	for(i=0; i<dev_list_cnt; i++) {
		if(!strcasecmp(dev_list[i].name, name) && dev_list[i].num_classes)
			return dev_list + i;
	}

	return NULL;
}

/* Process the sample array of length n.
 * If res_ang != 0, it will be filled with T evenly distanced angles.
 * If res_p != NULL, it will be filled with T evenly distanced pressures.
 * Return 0 on success, or 1 if the samples should be interpreted as a dot
 * rather than a longer sequence.
 */
static int process_samples(Sample sample[], int n, int *d_ang, int res_ang,
int *d_p, int res_p, int T) {
	int i,j,k,t;
	double dx, dy, r2, len;
	double d[n-1], r[n-1], ang[n-1], p[n-1];
	double min_p = MAXDOUBLE, max_p = -MAXDOUBLE;
	double scale_r, scale_p, oldang;
	double x, y;

	j = 0;
	k = 0;
	len = 0.0;
	for(i=1; i<n; i++) {
		dx = (double)(sample[i].x - sample[j].x);
		dy = (double)(sample[i].y - sample[j].y);
		if((r2 = dx*dx + dy*dy) > move_treshold) {
			r[k] = sqrt(r2);
			len += r[k];
			d[k] = len;
			ang[k] = atan2(dy, dx);
			p[k] = (double)sample[i].p;
			if(p[k] > max_p) max_p = p[k];
			if(p[k] < min_p) min_p = p[k];
			j = i;
			k++;
		}
	}

	if(k < 4) return 1;

	scale_r = 1.0 / len;
	for(i=0; i<k; i++) {
		r[i] *= scale_r;
		d[i] *= scale_r;
	}

	for(i=0; i<k; i++) ang[i] = (ang[i]+M_PI)*0.5*M_1_PI;

	if(max_p > min_p) {
		scale_p = 1.0 / (max_p - min_p);
		for(i=0; i<k; i++) p[i] = (p[i] - min_p) * scale_p;
	} else {
		for(i=0; i<k; i++) p[i] = 0.0;
	}

	x = 0.0;
	i = 1;
	for(t=0; t<T; t++) {
		while(x > d[i]) i++;
		if(res_ang) {
			if(ang[i-1] < 0.25 && ang[i] > 0.75) oldang = ang[i-1] + 1.0;
			else if(ang[i-1] > 0.75 && ang[i] < 0.25) oldang = ang[i-1] - 1.0;
			else oldang = ang[i-1];

			if(x < d[i-1]) y = ang[i-1]; else
			y = ang[i-1] + (ang[i]-oldang) * ((x-d[i-1]) / (d[i]-oldang));
			d_ang[t] = ((int)(y*(double)(res_ang))) % (unsigned)res_ang;
		}
		if(res_p) {
			if(x < d[i-1]) y = p[i-1]; else
			y = p[i-1] + (p[i]-p[i-1]) * ((x-d[i-1]) / (d[i]-d[i-1]));
			d_p[t] = (int)(y*(double)(res_p-1));
		}
		x += 1.0 / (double)T;
	}
/*
	printf("Sample:\n");
	for(i=0; i<k; i++) {
		printf("  %d: d = %g, ang = %g, p = %g\n",
			i, d[i], ang[i], p[i]);
	}
*/
	return 0;
}

/* Initialize XInput and start capturing data. Each input sequence will be
 * reported to the callback function cb. A NULL argument to the callback means
 * the input was a dot. T is the length of the sequences.
 * res[0] is the angular resolution, or 0 if angles are not to be reported.
 * res[1] is the pressure resolution, or 0 is pressure is not to be reported.
 * The sequence passed to the callback will contain as many variables as are
 * necessary, e.g. 2 if res[0] and res[1] are both non-zero.
 * Note that the callback is responsible to call sequence_destroy() to free up
 * the memory allocated by its argument.
 */
int xi_init(const char *dev_name, int(*cb)(Sequence*), int T, int res[]) {
	Display *disp;
	XDeviceInfo *dev_info;
	XDevice *dev;
	Window wnd;
	XEventClass event_list[INPUTEVENT_MAX];
	int event_list_cnt = 0;
	XEventClass cls;

	XEvent event;
	XAnyEvent *any;
	XDeviceMotionEvent *move;

	int pen_down = 0;
	Sample *buf = NULL;
	int buf_size = 0;
	int samples = 0;
	struct timeval tv;
	Sequence *seq;
	int retval;

	int idx = 0, i_ang = 0, i_p = 0;

	if(res[0]) i_ang = idx++;
	if(res[1]) i_p = idx++;

	disp = get_display();
	if((dev_info = get_device(disp, dev_name)) == NULL) return 1;

	wnd = XCreateWindow(
		disp,
		DefaultRootWindow(disp),
		0,0,10,10,		/* placement */
		0,				/* border width */
		0,				/* depth */
		InputOutput,
		CopyFromParent,
		0,
		NULL);
	XMapWindow(disp, wnd);
	dev = XOpenDevice(disp, dev_info->id);
	if(dev == NULL) {
		fprintf(stderr, "Unable to open input device.\n");
		XDestroyWindow(disp, wnd);
		goto shutdown;
	}

	DeviceButtonPress(dev, input_event[INPUTEVENT_BTN_PRESS], cls);
	if(cls) event_list[event_list_cnt++] = cls;
	DeviceButtonRelease(dev, input_event[INPUTEVENT_BTN_RELEASE], cls);
	if(cls) event_list[event_list_cnt++] = cls;
	DeviceMotionNotify(dev, input_event[INPUTEVENT_MOTION_NOTIFY], cls);
	if(cls) event_list[event_list_cnt++] = cls;

	XGrabDevice(disp, dev, wnd, 0, event_list_cnt, event_list,
		GrabModeAsync, GrabModeAsync, CurrentTime);

	for(;;) {
		XNextEvent(disp, &event);
		any = (XAnyEvent*) &event;

		if(any->type == input_event[INPUTEVENT_MOTION_NOTIFY]) {
			move = (XDeviceMotionEvent*)any;
			if(use_pressure) {
				if(move->axis_data[2] > pressure_treshold[0] && !pen_down) {
					pen_down = 1;
					buf_size = 16;
					samples = 0;
					buf = malloc(buf_size*sizeof(*buf));
				} else
				if(move->axis_data[2] < pressure_treshold[1] && pen_down) {
					pen_down = 0;
					seq = sequence_create(T, idx);
					if(!process_samples(buf, samples, seq->o[i_ang], res[0],
						seq->o[i_p], res[1], T)) {
						retval = cb(seq);
						/* sequence_destroy(seq); */
					} else {
						retval = cb(NULL);
						sequence_destroy(seq);
					}
					free(buf);
					if(retval) goto shutdown;
				}
			}
			if(pen_down) {
				gettimeofday(&tv, NULL);
				if(samples >= buf_size) {
					buf_size *= 2;
					buf = realloc(buf, buf_size*sizeof(*buf));
				}
				buf[samples].x = move->axis_data[0];
				buf[samples].y = move->axis_data[1];
				buf[samples].p = move->axis_data[2];
				buf[samples].tx = (short)(move->axis_data[3]&0xffff);
				buf[samples].ty = (short)(move->axis_data[4]&0xffff);
				buf[samples].t =
					(1.0e-6*(double)tv.tv_usec) + (double)tv.tv_sec;
				samples++;
			}
		} else if(!use_pressure) {
			if(any->type == input_event[INPUTEVENT_BTN_PRESS]) {
				pen_down = 1;
				buf_size = 16;
				samples = 0;
				buf = malloc(buf_size*sizeof(*buf));
			} else if(any->type == input_event[INPUTEVENT_BTN_RELEASE]) {
				/* TODO: redundant, try to merge with code above */
				pen_down = 0;
				seq = sequence_create(T, idx);
				if(!process_samples(buf, samples, seq->o[i_ang], res[0],
					seq->o[i_p], res[1], T)) {
					retval = cb(seq);
					/* sequence_destroy(seq); */
				} else {
					retval = cb(NULL);
					sequence_destroy(seq);
				}
				free(buf);
				if(retval) goto shutdown;
			}		
		}
	}

shutdown:;
	XUngrabDevice(disp, dev, CurrentTime);
	XFree(dev);
	XDestroyWindow(disp, wnd);
	XCloseDisplay(disp);

	return 1;
}

